@@ -25,12 +25,12 @@ class Mamba2Block(nn.Module):
         d_in_proj = 2 * self.d_inner + 2 * self.d_state + self.n_heads
         self.in_proj = nn.Linear(self.d_model, d_in_proj, bias=False, device=device)
 
-        cond_inner = self.d_inner + 2 * self.d_state
+        conv_dim = self.d_inner + 2 * self.d_state
         self.conv1d = nn.Conv1d(
-            in_channels=cond_inner,
-            out_channels=cond_inner,
+            in_channels=conv_dim,
+            out_channels=conv_dim,
             kernel_size=self.d_conv,
-            groups=cond_inner,
+            groups=conv_dim,
             padding=self.d_conv-1,
             device=device,
         )
@@ -278,45 +278,80 @@ class SeqLinear(nn.Module):
     """
     Linear layer with weights derived from a weighted sum of rank-1 sequential updates.
     """
-    def __init__(self, d_in, d_out, n_heads,
+    def __init__(self, d_model, n_heads,
+                 d_conv=4, d_state=None, d_inner=None,
                  device="cpu"):
         super().__init__()
-        self.d_in = d_in
-        self.d_out = d_out
+        self.d_model = d_model
+        self.d_state = d_state if d_state is not None else d_model
+        self.d_inner = d_inner if d_inner is not None else d_model
         self.n_heads = n_heads
         self.device = device
         
-        d_in_proj = d_in + d_out + self.n_heads
-        self.in_proj = nn.Linear(d_in, d_in_proj, bias=False, device=device)
+        self.alpha = (6 / (self.d_state + self.d_inner)) ** 0.5
+        d_in_proj = 2 * self.d_state + self.d_inner + 2 * self.n_heads
+        self.in_proj = nn.Linear(d_model, d_in_proj, bias=False, device=device)
+        self.conv = nn.Conv1d(
+            in_channels=d_in_proj,
+            out_channels=d_in_proj,
+            kernel_size=d_conv,
+            groups=d_in_proj,
+            padding=d_conv-1,
+            device=device,
+        )
         self.w_base = nn.Parameter(torch.empty(n_heads, device=device))
+        self.out_proj = nn.Linear(d_inner, d_model, bias=False, device=device)
         
         self._reset_parameters()
     
     def _reset_parameters(self):
         nn.init.xavier_uniform_(self.in_proj.weight)
+        nn.init.xavier_uniform_(self.conv.weight)
         nn.init.uniform_(self.w_base, -0.1, 0.1)
+        nn.init.xavier_uniform_(self.out_proj.weight)
+        
+        if self.conv.bias is not None:
+            nn.init.constant_(self.conv.bias, 0.)
     
     def bidirectional(self, x):
         bsz, src_len, d_model = x.size()
         
-        IOw = self.in_proj(x)
+        xIOw_f = self.conv(self.in_proj(x).transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len]
+        xIOw_b = self.conv(self.in_proj(x).flip(1).transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len]
         
-        I, O, w = torch.split(
-            IOw,
+        x_f, I_f, O_f, w_f, g_f = torch.split(
+            xIOw_f,
             [
-                self.d_in,
-                self.d_out,
+                self.d_state,
+                self.d_state,
+                self.d_inner,
+                self.n_heads,
+                self.n_heads,
+            ],
+            dim=-1
+        )
+        x_b, I_b, O_b, w_b, g_b = torch.split(
+            xIOw_b,
+            [
+                self.d_state,
+                self.d_state,
+                self.d_inner,
+                self.n_heads,
                 self.n_heads,
             ],
             dim=-1
         )
         
-        x = rearrange(x, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
-        I = rearrange(I, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
-        O = rearrange(O, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        x_f = rearrange(self.alpha * x_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        I_f = rearrange(I_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        O_f = rearrange(O_f, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        
+        x_b = rearrange(self.alpha * x_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        I_b = rearrange(I_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        O_b = rearrange(O_b, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
 
-        decay_f = F.silu(w * self.w_base).transpose(1, 2).unsqueeze(-2).repeat(1, 1, src_len, 1)
-        decay_b = decay_f.flip(-1)
+        decay_f = F.silu(w_f * self.w_base).transpose(1, 2).unsqueeze(-2).repeat(1, 1, src_len, 1)
+        decay_b = F.silu(w_b * self.w_base).transpose(1, 2).unsqueeze(-2).repeat(1, 1, src_len, 1)
         mask = torch.tril(torch.ones(src_len, src_len, dtype=torch.bool, device=self.device), diagonal=-1)
         decay_f = decay_f.masked_fill(~mask, 0.0)
         decay_b = decay_b.masked_fill(~mask, 0.0)
@@ -327,29 +362,33 @@ class SeqLinear(nn.Module):
         decay_b = decay_b.masked_fill(~mask, float('-inf'))
         decay_f = F.softmax(decay_f, dim=-1)
         decay_b = F.softmax(decay_b, dim=-1)
-        out_f = torch.einsum('bthd, bhts, bshd, bshe -> bthe', x, decay_f, I, O)
-        out_b = torch.einsum('bthd, bhts, bshd, bshe -> bthe', x, decay_b, I, O)
+        out_f = torch.einsum('bthd, bhts, bshd, bshe -> bthe', x_f, decay_f, I_f, O_f)
+        out_b = torch.einsum('bthd, bhts, bshd, bshe -> bthe', x_b, decay_b, I_b, O_b)
         
-        out = rearrange(out_f + out_b, 'b s h d -> b s (h d)')
+        out = out_f * F.sigmoid(g_f).unsqueeze(-1) + out_b * F.sigmoid(g_b).unsqueeze(-1)
+        out = rearrange(out, 'b s h d -> b s (h d)')
+        out = self.out_proj(out)
         return out
     
     def forward(self, x, mode="causal"):
         if mode == "bidirectional": return self.bidirectional(x)
         bsz, src_len, d_model = x.size()
 
-        IOw = self.in_proj(x)
+        xIOw = self.conv(self.in_proj(x).transpose(1, 2).contiguous()).transpose(1, 2)[:, :src_len]
         
-        I, O, w = torch.split(
-            IOw,
+        x, I, O, w, g = torch.split(
+            xIOw,
             [
-                self.d_in,
-                self.d_out,
+                self.d_state,
+                self.d_state,
+                self.d_inner,
+                self.n_heads,
                 self.n_heads,
             ],
             dim=-1
         )
         
-        x = rearrange(x, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
+        x = rearrange(self.alpha * x, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         I = rearrange(I, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         O = rearrange(O, 'b s (h d) -> b s h d', h=self.n_heads).contiguous()
         
@@ -366,7 +405,9 @@ class SeqLinear(nn.Module):
             decay = F.softmax(-F.silu(w * self.w_base), dim=1)
             out = torch.einsum('bthd, bsh, bshd, bshe -> bthe', x, decay, I, O)
 
+        out = out * F.sigmoid(g).unsqueeze(-1)
         out = rearrange(out, 'b s h d -> b s (h d)')
+        out = self.out_proj(out)
         return out
 
 class SeqMLP(nn.Module):
@@ -399,13 +440,7 @@ class SeqMLP(nn.Module):
         self.layers = nn.ModuleList([
             nn.ModuleDict(
                 dict(
-                    proj1=SeqLinear(
-                        d_in=emb_dim,
-                        d_out=2*emb_dim,
-                        n_heads=n_heads,
-                        device=device
-                    ),
-                    proj2=SeqLinear(
+                    mixer=SeqLinear(
                         d_in=emb_dim,
                         d_out=emb_dim,
                         n_heads=n_heads,
@@ -434,10 +469,7 @@ class SeqMLP(nn.Module):
             x = x + self.abs_pos_encoding(pos)
         for layer in self.layers:
             x_norm = layer.norm(x)
-            yg = layer.proj1(x_norm, mode=self.mode)
-            y, g = torch.chunk(yg, 2, dim=-1)
-            y = F.silu(y) * g
-            y = layer.proj2(y, mode=self.mode)
+            y = layer.mixer(x_norm, mode=self.mode)
             x = x + layer.dropout(y)
         x = self.norm_f(x)
         x = self.out_proj(x)
